#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "skiller.h"
#include "usb.h"
#include "utils.h"

#define PACKETS_MAX 255

skillerctl_packet_t packets[255] = {};
size_t count = 0;

struct color_stop {
    double   value;
    int16_t r;
    int16_t g;
    int16_t b;
} stops[] = {
    -INFINITY, 0x00, 0x00, 0x00,
    0/6., 0x00, 0x00, 0x00,
    1/6., 0x00, 0x00, 0xFF,
    2/6., 0x00, 0xFF, 0xFF,
    3/6., 0x00, 0xFF, 0x00,
    4/6., 0xFF, 0xFF, 0x00,
    5/6., 0xFF, 0x00, 0x00,
    6/6., 0xFF, 0xFF, 0xFF,
    INFINITY, 0xFF, 0xFF, 0xFF
};

int clamp(int min, int value, int max) {
    return value > max ? max : (value < min ? min : value);
}

uint32_t interpolate(double color) {
    struct color_stop *top = stops;
    struct color_stop *bottom;

    while (top->value < color) top++;

    bottom = top - 1;

    double d = (color - bottom->value) / (top->value - bottom->value);

    return (clamp(0x00, bottom->r + (top->r - bottom->r)  * d, 0xFF) & 0xFF)
         + ((clamp(0x00, bottom->g + (top->g - bottom->g) * d, 0xFF) & 0xFF) << 8)
         + ((clamp(0x00, bottom->b + (top->b - bottom->b) * d, 0xFF) & 0xFF) << 16)
         ;
}

uint16_t mapping[] = {
    ['`']   = SK_KEY_TILDE,
    ['~']   = SK_KEY_TILDE,
    ['1']   = SK_KEY_1,
    ['!']   = SK_KEY_1,
    ['2']   = SK_KEY_2,
    ['@']   = SK_KEY_2,
    ['3']   = SK_KEY_3,
    ['#']   = SK_KEY_3,
    ['4']   = SK_KEY_4,
    ['$']   = SK_KEY_4,
    ['5']   = SK_KEY_5,
    ['%']   = SK_KEY_5,
    ['6']   = SK_KEY_6,
    ['^']   = SK_KEY_6,
    ['7']   = SK_KEY_7,
    ['&']   = SK_KEY_7,
    ['8']   = SK_KEY_8,
    ['*']   = SK_KEY_8,
    ['9']   = SK_KEY_9,
    ['(']   = SK_KEY_9,
    ['0']   = SK_KEY_0,
    [')']   = SK_KEY_0,
    ['-']   = SK_KEY_MINUS,
    ['_']   = SK_KEY_MINUS,
    ['=']   = SK_KEY_EQUALS,
    ['+']   = SK_KEY_EQUALS,
    ['q']   = SK_KEY_Q,
    ['Q']   = SK_KEY_Q,
    ['w']   = SK_KEY_W,
    ['W']   = SK_KEY_W,
    ['e']   = SK_KEY_E,
    ['E']   = SK_KEY_E,
    ['r']   = SK_KEY_R,
    ['R']   = SK_KEY_R,
    ['t']   = SK_KEY_T,
    ['T']   = SK_KEY_T,
    ['y']   = SK_KEY_Y,
    ['Y']   = SK_KEY_Y,
    ['u']   = SK_KEY_U,
    ['U']   = SK_KEY_U,
    ['i']   = SK_KEY_I,
    ['I']   = SK_KEY_I,
    ['o']   = SK_KEY_O,
    ['O']   = SK_KEY_O,
    ['p']   = SK_KEY_P,
    ['P']   = SK_KEY_P,
    ['[']   = SK_KEY_OPEN_BRACKET,
    ['{']   = SK_KEY_OPEN_BRACKET,
    [']']   = SK_KEY_CLOSE_BRACKET,
    ['}']   = SK_KEY_CLOSE_BRACKET,
    ['\\']  = SK_KEY_BACKSLASH,
    ['|']   = SK_KEY_BACKSLASH,
    ['a']   = SK_KEY_A,
    ['A']   = SK_KEY_A,
    ['s']   = SK_KEY_S,
    ['S']   = SK_KEY_S,
    ['d']   = SK_KEY_D,
    ['D']   = SK_KEY_D,
    ['f']   = SK_KEY_F,
    ['F']   = SK_KEY_F,
    ['g']   = SK_KEY_G,
    ['G']   = SK_KEY_G,
    ['h']   = SK_KEY_H,
    ['H']   = SK_KEY_H,
    ['j']   = SK_KEY_J,
    ['J']   = SK_KEY_J,
    ['k']   = SK_KEY_K,
    ['K']   = SK_KEY_K,
    ['l']   = SK_KEY_L,
    ['L']   = SK_KEY_L,
    [';']   = SK_KEY_COLON,
    [':']   = SK_KEY_COLON,
    ['\'']  = SK_KEY_QUOTE,
    ['"']   = SK_KEY_QUOTE,
    ['\n']  = SK_KEY_ENTER,
    [' ']   = SK_KEY_SPACE,
    ['z']   = SK_KEY_Z,
    ['Z']   = SK_KEY_Z,
    ['x']   = SK_KEY_X,
    ['X']   = SK_KEY_X,
    ['c']   = SK_KEY_C,
    ['C']   = SK_KEY_C,
    ['v']   = SK_KEY_V,
    ['V']   = SK_KEY_V,
    ['b']   = SK_KEY_B,
    ['B']   = SK_KEY_B,
    ['n']   = SK_KEY_N,
    ['N']   = SK_KEY_N,
    ['m']   = SK_KEY_M,
    ['M']   = SK_KEY_M,
    [',']   = SK_KEY_COMMA,
    ['<']   = SK_KEY_COMMA,
    ['.']   = SK_KEY_DOT,
    ['>']   = SK_KEY_DOT,
    ['/']   = SK_KEY_SLASH,
    ['?']   = SK_KEY_SLASH,
};

double heatmap[SK_LIMIT_KEY] = {};

void normalize_heatmap()
{
    double max = 1.0, min = 0.0;
    for (int i = 0; i < SK_LIMIT_KEY; i += 3) {
        max = max > heatmap[i] ? max : heatmap[i];
        min = min >= 1.0 && min < heatmap[i] ? min : heatmap[i]; 
    }

    for (int i = 0; i < SK_LIMIT_KEY; i += 3) {
        heatmap[i] /= (max - min);
    }
}

void make_row(skillerctl_keyaddr_t low, skillerctl_keyaddr_t high) {
    uint32_t colors[18] = {};

    unsigned j = 0;
    for (skillerctl_keyaddr_t i = low; i <= high; i += 3, j++) {
        printf("Key: %s heat: %lf color: #%06x\n", skillerctl_keystr(i), heatmap[i], interpolate(heatmap[i]));
        colors[j] = interpolate(heatmap[i]);
    }

    skillerctl_color_batch_command(packets + count++, low, colors, (high - low) / 3 + 1);
}

void make_packets()
{
    count = 0;
    skillerctl_disable_led_command(packets + count++);
    make_row(SK_KEY_TILDE, SK_KEY_EQUALS);
    make_row(SK_KEY_Q, SK_KEY_BACKSLASH);
    make_row(SK_KEY_A, SK_KEY_ENTER);
    make_row(SK_KEY_Z, SK_KEY_SLASH);
    skillerctl_color_command(packets + count++, SK_KEY_SPACE, interpolate(heatmap[SK_KEY_SPACE]));
    skillerctl_enable_led_command(packets + count++);
}

void handle_input(int argc, char **argv)
{
    char c = getc(stdin);
    skillerctl_keyaddr_t key;

    while (c != EOF) {
        if ((key = mapping[c])) {
            heatmap[key] += 1.;
        }

        c = getc(stdin);
    }

    normalize_heatmap();
    make_packets();
}

void handle_logic(libusb_device *skiller)
{
    libusb_device_handle *handle;
    skillerctl_packet_t packet;

    if (skiller == NULL) {
        printf("Cannot found compatibile keyboard, is it connected?\n");
        return;
    }

    if (libusb_open(skiller, &handle) != LIBUSB_SUCCESS) {
        printf("Cannot open device, try with sudo.\n");
        return;
    }

    libusb_detach_kernel_driver(handle, 1);
    libusb_claim_interface(handle, 1);

    for (int i = 0; i < count; i++) {
        skillerctl_send_command(handle, packets + i);
    }

    libusb_release_interface(handle, 1);
    libusb_attach_kernel_driver(handle, 1);
    libusb_close(handle);
} 

int main(int argc, char **argv) 
{
    libusb_device **list;

    libusb_init(NULL);
    size_t count = libusb_get_device_list(NULL, &list);
    libusb_device *skiller = skillerctl_find_device(list, count);

    handle_input(argc, argv);
    handle_logic(skiller);

    libusb_free_device_list(list, 1);
    libusb_exit(NULL);
}
