#ifndef SKILLER_H
#define SKILLER_H

#include <inttypes.h>
#include <stdlib.h>
#include "keys.h"

#define SK_MAX_PAYLOAD 56
#define SK_PACKET_SIZE 64

#define SK_MAGIC            0x04

#define SK_CMD_DISABLE_LED  0x01 // Disables key backlight, takes some time 
#define SK_CMD_ENABLE_LED   0x02 // Enables key backlight, takes some time
#define SK_CMD_SET_PROP     0x06 // Changes keyboard mode or mode settings
#define SK_CMD_MAP_KEYS     0x08 // Writes key mapping to controller 
#define SK_CMD_WRITE_MACRO  0x0A // Saves macros in controller memory
#define SK_CMD_SET_COLOR    0x11 // Changes LED RGB values in key given keys, or key range.

#define SK_PROP_MODE        0x00 // Changes backlight program (pulsing, permamant, custom, etc.) 
#define SK_PROP_BRIGHTNESS  0x01 // Sets brihgtness level 
#define SK_PROP_SPEED       0x02 // Sets animation speed 
#define SK_PROP_DIRECTION   0x03 // Sets animation direction 
#define SK_PROP_RGB         0x04 // LGBT promotion 
#define SK_PROP_COLOR       0x05 // Base color 
#define SK_PROP_POLLING     0x0F // Polling rate

#define SK_MODE_WAVE        0x01 // Wave 
#define SK_MODE_DRIFT       0x02 // Drift 
#define SK_MODE_SWIRL       0x03 // Swirl 
#define SK_MODE_CHANGE      0x04 // color change 
#define SK_MODE_PULSATING   0x05 // pulsating 
#define SK_MODE_PERMAMENT   0x06 // permament 
#define SK_MODE_TRIGGER     0x07 // Trigger 
#define SK_MODE_EXPLOSION   0x08 // Explosion 
#define SK_MODE_BURST       0x09 // Gamma Ray Burst 
#define SK_MODE_CHAOS       0x0A // Chaos 
#define SK_MODE_COSMIC      0x0B // Cosmic 
#define SK_MODE_GRADIENT    0x0C // Gradient 
#define SK_MODE_TIDE        0x0D // Tide 
#define SK_MODE_LIGHTING    0x0E // Ball Lighting 
#define SK_MODE_MATRIX      0x0F // Matrix 
#define SK_MODE_RICOCHET    0x10 // Ricochet 
#define SK_MODE_RIPPLE      0x12 // Ripple 
#define SK_MODE_CUSTOM      0x14 // Custom

typedef uint16_t skillerctl_property_t;
typedef uint8_t  skillerctl_mode_t;

typedef struct {
    uint8_t  magic;
    uint16_t checksum;
    uint8_t  command;
    uint16_t address;
    uint8_t  length;
    uint8_t  payload[SK_MAX_PAYLOAD];
} skillerctl_packet_t;

typedef struct {
    uint8_t r;
    uint8_t g;
    uint8_t b;
} skillerctl_color_t;

typedef struct {
    skillerctl_color_t color[SK_MAX_PAYLOAD / 3];
} skillerctl_payload_color_t;

void skillerctl_make_packet(skillerctl_packet_t *packet, uint8_t command, uint16_t address, void* payload, uint8_t size);

void skillerctl_color_command(skillerctl_packet_t *packet, skillerctl_keyaddr_t address, uint32_t color);
void skillerctl_color_batch_command(skillerctl_packet_t *packet, skillerctl_keyaddr_t start, uint32_t *colors, size_t n);

void skillerctl_enable_led_command(skillerctl_packet_t *packet);
void skillerctl_disable_led_command(skillerctl_packet_t *packet);

void skillerctl_set_mode_command(skillerctl_packet_t *packet, skillerctl_mode_t mode);
void skillerctl_set_brightness_command(skillerctl_packet_t *packet, uint8_t brightness);
void skillerctl_set_direction_command(skillerctl_packet_t *packet, uint8_t direction);
void skillerctl_set_rgb_command(skillerctl_packet_t *packet, uint8_t rgb);
void skillerctl_set_base_color_command(skillerctl_packet_t *packet, uint32_t color);
void skillerctl_set_speed_command(skillerctl_packet_t *packet, uint8_t speed);
void skillerctl_set_polling_command(skillerctl_packet_t *packet, uint8_t rate);

skillerctl_packet_t* skillerctl_unpack(skillerctl_packet_t *packet, uint8_t *data, size_t length);
void*                skillerctl_pack(skillerctl_packet_t *packet, uint8_t *data);

uint16_t skillectl_packet_checksum(skillerctl_packet_t *packet);

char*             skillerctl_modestr(skillerctl_mode_t mode);
skillerctl_mode_t skillerctl_strmode(const char *mode);

#endif /* SKILLER_H */

